module nonfact
  use helper
  use constants
  use runge_kutta
  use incl_parameters
!  use grids
  implicit none

  public f1
  public f2
  public chinf

  private f1_integrand
  private f2_integrand
  private runge_kutta_dx

  real(dp),save :: q1sq_rk,q2sq_rk,qHsq_rk
  real(dp), save :: MVsq ! Mass of vector boson, ie either MW or MZ
contains

!  subroutine generate_f1_f2_grid(mboson,qsqmin,qsqmax,npoints,filename)
!    real(dp),intent(in) :: mboson,qsqmin,qsqmax
!    integer, intent(in) :: npoints
!    character(100), intent(in) :: filename
!    character(100) :: f1name, f2name
!    integer :: order
!    logical :: isLogx, isLogy,file_exists
!    MV = mboson
!    order = 0
!    isLogx = .true.
!    isLogy = .true.
!    f1name = "f1_" //trim(filename)//".grid"
!    inquire(FILE=trim(f1name), EXIST=file_exists)
!    ! First initialise f1
!    if(file_exists) then
!       call initGrid2D(f1grid,f1,qsqmin,qsqmax,npoints,qsqmin,qsqmax,npoints,order,isLogx,&
!            &isLogy,.true.,.false.,f1name)
!    else
!       call initGrid2D(f1grid,f1,qsqmin,qsqmax,npoints,qsqmin,qsqmax,npoints,order,isLogx,&
!            &isLogy,.false.,.true.,f1name)
!    endif
!  end subroutine generate_f1_f2_grid
  function f1(q1sq,q2sq,qHsq,MV) result(res)
    real(dp), intent(in) :: q1sq,q2sq,qHsq,MV
    real(dp) :: res
    integer :: iterations

    ! Save input for runge kutta routine
    q1sq_rk = q1sq
    q2sq_rk = q2sq
    qHsq_rk = qHsq
    MVsq    = MV**2
    iterations = niter
!    iterations = 10000

    call runge_kutta_dx(f1_integrand,zero,one,zero,res,iterations)
  end function f1

  function f1_analytic(q1sq,q2sq,qHsq,MV) result(res)
    real(dp), intent(in) :: q1sq,q2sq,qHsq,MV
    real(dp) :: res, delta1, delta2
    integer :: iterations

    MVsq    = MV**2
    delta1 = q1sq + MVsq
    delta2 = q2sq + MVsq

    res = & 
     & -(((-MVsq - q1sq)* &
     &       (q2sq*(q1sq - q2sq + qHsq) + & 
     &         MVsq*(-q1sq + q2sq + qHsq))* &
     &       Log(delta2/delta1))/ &
     &     (-(MVsq*(q1sq - q2sq)**2) + delta1*delta2*qHsq) &
     &     ) + 2*Log(1 + q1sq/Mv**2) + &
     &  (delta2*(-MVsq - q1sq)* &
     &     (-2*MVsq + q1sq + q2sq - qHsq)* &
     &     Sqrt(qHsq/(4*MVsq + qHsq))* &
     &     Log((-Sqrt(qHsq) + Sqrt(4*MVsq + qHsq))/ &
     &       (Sqrt(qHsq) + Sqrt(4*MVsq + qHsq))))/ &
     &   (-(MVsq*(q1sq - q2sq)**2) + delta1*delta2*qHsq)
  end function f1_analytic
  ! Eq. 9 + 10 of 1906.10899 
  function f1_integrand(x,y) result(res)
    real(dp), intent(in) :: x,y
    real(dp) :: res
    ! Internal
    real(dp) :: delta1,delta2,r1,r2,r12

    r1 = q1sq_rk * x + q2sq_rk * (one - x) - qHsq_rk * x *(one - x)
    r2 = qHsq_rk * x * (one - x) + MVsq
    r12 = r1 + r2
    delta1 = q1sq_rk + MVsq
    delta2 = q2sq_rk + MVsq

    res = delta1*delta2/r12**2 * (log(r12**2/(r2*MVsq)) + (r1 - r2)/r2)
  end function f1_integrand

  function f2(q1sq,q2sq,qHsq,MV) result(res)
    real(dp), intent(in) :: q1sq,q2sq,qHsq,MV
    real(dp) :: res
    integer :: iterations

    ! Save input for runge kutta routine
    q1sq_rk = q1sq
    q2sq_rk = q2sq
    qHsq_rk = qHsq
    MVsq    = MV**2
    iterations = niter

    call runge_kutta_dx(f2_integrand,zero,one,zero,res,iterations)
  end function f2
  ! Eq. 9 + 10 of 1906.10899 
  function f2_integrand(x,y) result(res)
    real(dp), intent(in) :: x,y
    real(dp) :: res
    ! Internal
    real(dp) :: delta1,delta2,r1,r2,r12
    complex(dp) :: HPL2
    complex(dp) :: z
    integer :: n1,n2
    real(dp) :: Li2

    ! This returns Li2
    n1 = 0
    n2 = 1
!    print*, 'HPL2', HPL2(n1,n2,z)
    
    r1 = q1sq_rk * x + q2sq_rk * (one - x) - qHsq_rk * x *(one - x)
    r2 = qHsq_rk * x * (one - x) + MVsq
    r12 = r1 + r2
    delta1 = q1sq_rk + MVsq
    delta2 = q2sq_rk + MVsq

    z = r1/r12
    Li2 = real(HPL2(n1,n2,z))
    res = delta1*delta2/r12**2 * ( &
         & (log(r12**2/(r2*MVsq)) + (r1 - r2)/r2)**2 &
         & - log(r12/r2)**2 - two*r12/r2*log(r12/r2) &
         & - two*Li2 - ((r1-r2)/r2)**2 + two*zeta2)
  end function f2_integrand

  function chinf(qT1,qT2,MV) result(res)
    real(dp), intent(in) :: qT1(1:2), qT2(1:2),MV
    real(dp) :: qTH(1:2), ptH, pt1, pt2, res

    qTH = - qT1 - qT2
    ptH = qTH(1)**2 + qTH(2)**2
    pt1 = qT1(1)**2 + qT1(2)**2
    pt2 = qT2(1)**2 + qT2(2)**2

    !    res = f1(pt1,pt2,ptH,MV)**2 - f2(pt1,pt2,ptH,MV)
    res = zero
    if(oneloop_on) res = res + f1_analytic(pt1,pt2,ptH,MV)**2 
    if(twoloop_on) res = res - f2(pt1,pt2,ptH,MV)
  end function chinf

  subroutine runge_kutta_dx(f,x0,x1,y0,integral,n_iter)
      
      real(dp), intent(in) :: x0,x1,y0
      integer, intent(in) :: n_iter
      real(dp), intent(out) :: integral
      real(dp) :: k(1:4), delx_6, delx_2
      real(dp) :: x(1:4), delx
      integer :: iter
      logical :: debug = .false.
      interface
         function f(x,y) result(res)
           use helper
           implicit none
           real(dp), intent(in) :: x, y
           real(dp) :: res
         end function f
      end interface

      if(n_iter.lt.1) then
         print*, n_iter
         stop 'n_iter should be positive integer'
      endif
      
      integral = y0 ! Initial value
      delx = (x1 - x0) / n_iter ! Step size
      delx_6 = delx/six
      delx_2 = delx/two
      k = zero
      x(1) = x0
      x(2) = x0 + half * delx
      !      x(3) = x0 + half * delx
      x(3) = x(2)
      x(4) = x0 + delx

      if(debug) then
         print*, 'x0        ', x0
         print*, 'x1        ', x1
         print*, 'y0        ', y0
         print*, 'n_iter    ', n_iter
         print*, 'x(1:4)    ', x
         print*, 'delx      ', delx
         print*, 'delx_2    ', delx_2
         print*, 'delx_6    ', delx_6
         print*, ''
      endif

      do iter = 1, n_iter
         k(1) = f(x(1),integral)
         k(2) = f(x(2),integral + delx_2 * k(1))
         k(3) = f(x(3),integral + delx_2 * k(2))
         k(4) = f(x(4),integral + delx * k(3))

         integral = integral + &
              & delx_6 * (k(1) + two*k(2) + two*k(3) + k(4))
         x(1:4) = x(1:4) + delx
         if(debug) then
            print*, 'x0        ', x0
            print*, 'x1        ', x1
            print*, 'y0        ', y0
            print*, 'iter      ', iter
            print*, 'x(1:4)    ', x
            print*, 'delx      ', delx
            print*, 'delx_2    ', delx_2
            print*, 'delx_6    ', delx_6
            print*, 'k(1:4)    ', k
            print*, 'integral  ', integral
            if(integral.ne.integral) stop
         endif
      enddo

    end subroutine runge_kutta_dx
end module nonfact
